/*

Copyright (c) 2005-2016, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <iostream>
#include <string>

#include "ExecutableSupport.hpp"
#include "Exception.hpp"
#include "PetscTools.hpp"
#include "CommandLineArguments.hpp"
#include "PetscException.hpp"
#include "TrianglesMeshReader.hpp"
#include "SimpleImpedanceProblem.hpp"
#include "ImpedancePostProcessor.hpp"
int main(int argc, char *argv[])
{
    // This sets up PETSc and prints out copyright information, etc.
    ExecutableSupport::StandardStartup(&argc, &argv);

    int exit_code = ExecutableSupport::EXIT_OK;

    // You should put all the main code within a try-catch, to ensure that
    // you clean up PETSc before quitting.
    try
    {
        if (PetscTools::IsParallel())
        {
            ExecutableSupport::PrintError("This executable does not yet run in parallel", true);
            exit_code = ExecutableSupport::EXIT_ERROR;
        }
        else if (argc<2)
        {
            ExecutableSupport::PrintError("Usage: ./ChasteLungImpedance.sh <options> path_to_mesh\n Options: -scale_radii <0.0-1.0>\tScale all airway radii (defaults to 0.7)", true);
            exit_code = ExecutableSupport::EXIT_BAD_ARGUMENTS;
            /**
             * \todo Things which could be configured at this stage
             *  * Mesh in millimetres -- problem.SetMeshInMilliMetres()
             *  * Frequencies         -- std::vector<double> test_frequencies;
             *  * Elastance           -- problem.SetElastance(5.8*98.0665*1e3);
             *  * Output location     -- std::cout << "R5...
             */
        }
        else
        {
            /* Get final argument and remove the suffix (if necessary) */
            std::string filename = argv[argc-1];
            size_t last_index = filename.find_last_of(".");
            std::string pathname = filename;
            // If last_index == 0 then this looks like a relative path with no suffix.
            if (last_index != 0)
            {
                pathname = filename.substr(0, last_index);

            }

            /* Get scaling factor from the command line arguments */
            double scale_radii = 0.7;
            if (CommandLineArguments::Instance()->OptionExists("-scale_radii"))
            {
                scale_radii = CommandLineArguments::Instance()->GetDoubleCorrespondingToOption("-scale_radii");
                if (scale_radii <= 0.0 || scale_radii > 1.0)
                {
                    EXCEPTION("The scaling should be between 0.0 and 1.0");
                }
            }

            /* First, we load up a mesh containing the centre lines and radii of the a complete conducting airway tree.
             * The mesh will typically have been developed using a combination of computed tomography (CT) image segmentation
             * and algorithmic airway generation.
             */
            TetrahedralMesh<1,3> mesh;
            ExecutableSupport::Print("Loading mesh files with base file name: "+ pathname);
            TrianglesMeshReader<1,3> mesh_reader(pathname);
            mesh.ConstructFromMeshReader(mesh_reader);



            /* Note that the mesh defined above was developed using a CT scan taken at full inspiration. Impedance is more
             * commonly recorded during tidal breathing. Here we use a simple scaling to bring the airway radii down into the
             * tidal breathing range.
             */
            if (scale_radii < 1.0)
            {
                std::stringstream sstream;

                sstream << "Scaling all airway radii by a factor of "<< scale_radii;
                ExecutableSupport::Print(sstream.str());
                for (TetrahedralMesh<1,3>::NodeIterator node_iter = mesh.GetNodeIteratorBegin();
                        node_iter != mesh.GetNodeIteratorEnd();
                        ++node_iter)
                {
                    node_iter->rGetNodeAttributes()[0] *= scale_radii;
                }
            }
            else
            {
                ExecutableSupport::Print("Not scaling airway radii (scale_radii = 1.0)");
            }

            /* Setup a !SimpleImpedanceProblem and tell it that the given mesh is defined in millimetres
             */
            SimpleImpedanceProblem problem(mesh, 0u);
            problem.SetMeshInMilliMetres();


            /* This vector lists the input frequencies at which to calculate impedance. They must be
             * monotonically increasing.
             */
            std::vector<double> test_frequencies;
            test_frequencies.push_back(1.0);
            test_frequencies.push_back(2.0);
            test_frequencies.push_back(3.0);
            test_frequencies.push_back(5.0);
            test_frequencies.push_back(10.0);
            test_frequencies.push_back(20.0);
            test_frequencies.push_back(30.0);
            problem.SetFrequencies(test_frequencies);               //Set & get frequencies for coverage

            /* The simple impedance model defines a linear spring at each terminal of the airway tree.
             * This method allows us to set the elastance of the whole lung (in Pa/m^3). This elastance
             * is then evenly distributed over the terminals.
             */
            problem.SetElastance(5.8*98.0665*1e3);

            /* Calculates the impedance at the given frequencies*/
            problem.Solve();

            /* Get the calculated impedances. The impedance at each frequency is
             * made up of a real component (the resistance) and a complex component
             * (the elastance).
             */
            std::vector<std::complex<double> > impedances = problem.rGetImpedances();

            /* The impedances calculated above could at this stage be written to a file
             * and plotted. Instead, we make use of !ImpedancePostProcessor to calculate
             * a number of common clinical summary statistics from the data.
             */
            ImpedancePostProcessor processor(test_frequencies, impedances);

            std::cout << "\n";
            std::cout << "R5       = " << real(impedances[3])*1e-6 << " kPa.s.L^-1" << std::endl;
            std::cout << "R20      = " << real(impedances[5])*1e-6 << " kPa.s.L^-1" << std::endl;
            std::cout << "R5 - R20 = " << processor.GetR5MinusR20()*1e-6 << " kPa.s.L^-1" << std::endl;
            std::cout << "Rrs      = " << processor.GetRrs()*1e-6 << " kPa.s.L^-1" << std::endl;
            std::cout << "\n";
            std::cout << "X5       = " << imag(impedances[5])*1e-6 << " kPa.s.L^-1" << std::endl;
            std::cout << "Fres     = " << processor.GetResonantFrequency() << " Hz " << std::endl;
            std::cout << "Ax       = " << processor.GetAx()*1e-6 << " kPa.L^-1" << std::endl;
            std::cout << "\n";
        }
    }
    catch (const Exception& e)
    {
        ExecutableSupport::PrintError(e.GetMessage());
        exit_code = ExecutableSupport::EXIT_ERROR;
    }

    // Optional - write the machine info to file.
    ExecutableSupport::WriteMachineInfoFile("machine_info");

    // End by finalizing PETSc, and returning a suitable exit code.
    // 0 means 'no error'
    ExecutableSupport::FinalizePetsc();
    return exit_code;
}
